#include "ModelSearch.h"

#include <math.h>
#include <assert.h>
#include "pq.h"

/* lame hack globals */
extern int   gMaxParentsPerNode;
extern long  gMaxParameterCount;
extern long  gMaxBytesPerModel;
extern int   gLimitBytes;
extern float gDelta;
extern int   gP0Multiplier;
extern long  gNumBoundsUsed;
extern float gTau;
extern int   gAllowRemove;
extern int   gLazyCycleCheck;

extern long  gNumByTie;
extern long  gNumByWin;
extern long  gMinSamplesInDecision;
extern long  gMaxSamplesInDecision;
extern long  gTotalSamplesInDecision;
extern long  gNumZeroSamplesInDecision;
extern long  gNumCurrentInTie;
extern long  gNumCurrentByWin;
extern long  gNumCurrentByDefault;
extern long  gNumByCycleConflict;
extern long  gNumByParameterConflict;
extern long  gNumByParentLimit;
extern long  gNumByMemoryLimit;
extern long  gNumByParameterLimit;
extern long  gNumByChangeLimit;
extern long  gNumRemoved;
extern long  gNumAdded;
extern long  gNumUsedWholeDB;


typedef struct BNUserData_ {
   BeliefNetNode changedOne;
   int changedParentID;

   /* -1 for removal, 0 for nothing, 1 for reverse, 2 for add */
   int changeComplexity;

   BeliefNet currentNet;
   int myTmpInternalData;
} BNUserDataStruct, *BNUserData;


typedef struct BNNodeUserData_ {
   double         avgDataLL;
//   double         score;
//   double         upperBound;
//   double         lowerBound;
   double         p0;
   int            isChangedFromCurrent;
   struct BNNodeUserData_ *current; // null if this is a current node
} BNNodeUserDataStruct, *BNNodeUserData;

static void _FreeUserData(BNUserData netData) {

   if(netData->changedOne) {
      MFreePtr(BNNodeGetUserData(netData->changedOne));
      BNNodeFree(netData->changedOne);
   }

   MFreePtr(netData);
}


int _BNHasCycle(BNUserData netData) {
   //BeliefNet newBN;
   BeliefNetNode changedNode, parentNode;
   int hasCycle;

   if(netData->changeComplexity == -1 || netData->changeComplexity == 0) {
      hasCycle = BNHasCycle(netData->currentNet);
   } else {
      changedNode = BNGetNodeByID(netData->currentNet,
                                   BNNodeGetID(netData->changedOne));
      parentNode = BNGetNodeByID(netData->currentNet,
                                          netData->changedParentID);

      if(BNNodeGetNumChildren(changedNode) == 0 || 
             BNNodeGetNumParents(parentNode) == 0) {
         hasCycle = 0;
      } else {
         //newBN = BNCloneNoCPTs(netData->currentNet);

        //changedNode = BNGetNodeByID(newBN, BNNodeGetID(netData->changedOne));
        //parentNode = BNGetNodeByID(newBN, netData->changedParentID);

         if(netData->changeComplexity == 2) {
            BNFlushStructureCache(netData->currentNet);
            BNNodeAddParent(changedNode, parentNode);
            hasCycle = BNHasCycle(netData->currentNet);
            BNNodeRemoveParent(changedNode, 
                       BNNodeLookupParentIndex(changedNode, parentNode));
            BNFlushStructureCache(netData->currentNet);
         } else {
            /* must be a remove, but that is special cased elsewhere
                    and can't cause a cycle */
            //BNNodeRemoveParent(changedNode, 
            //     BNNodeLookupParentIndex(changedNode, parentNode));
            hasCycle = 0;
         }
         //hasCycle = BNHasCycle(newBN);
         //BNFree(newBN);
      }
   }

   return hasCycle;
}

BeliefNetNode _BNNodeCloneNoCPT(BeliefNetNode bnn, BeliefNet newBN);
static BNUserData _InitUserData(int nodeID, int doChange, int changedParentID,
                                 int isAdd, BeliefNet current) {
   BNNodeUserData data;
   BNUserData netData;
   BeliefNetNode changedBnn, currentBnn;
   int changeComplexity = 0;

   if(doChange) {
      changedBnn = _BNNodeCloneNoCPT(BNGetNodeByID(current, nodeID),
                                                current);
      data = MNewPtr(sizeof(BNNodeUserDataStruct));
      BNNodeSetUserData(changedBnn, data);

      data->avgDataLL = 0;

      data->isChangedFromCurrent = 1;
      data->p0 = 1;

      currentBnn = BNGetNodeByID(current, nodeID);
      data->current = BNNodeGetUserData(currentBnn);

      if(isAdd) {
         changeComplexity = 2;
         /* HERE break BN ADT right here */
         VLAppend(changedBnn->parentIDs, (void *)changedParentID);
         changedBnn->numParentCombinations = -1;
         BNNodeInitCPT(changedBnn);
      } else {
         /* remove */
         changeComplexity = -1;
         /* HERE break BN ADT right here */
         VLRemove(changedBnn->parentIDs, 
                     BNNodeLookupParentIndexByID(changedBnn, changedParentID));
         changedBnn->numParentCombinations = -1;
         BNNodeInitCPT(changedBnn);
      }

      /* HERE HERE Heuristic hack */
      //BNNodeSmoothProbabilities(changedBnn, 1.0 / 
      //                      (float)BNNodeGetNumValues(changedBnn));
   } else {
      changedBnn = 0;
   }

   netData = MNewPtr(sizeof(BNUserDataStruct));
   netData->changedOne = changedBnn;
   netData->changedParentID = changedParentID;
   netData->changeComplexity = changeComplexity;
   netData->currentNet = current;

   return netData;
}

static void _InitCurrentUserData(BeliefNet current) {
   int i;
   BeliefNetNode bnn;
   BNNodeUserData data;

   if(BNGetUserData(current) == 0) {
      for(i = 0 ; i < BNGetNumNodes(current) ; i++) {
         bnn = BNGetNodeByID(current, i);
      
         data = MNewPtr(sizeof(BNNodeUserDataStruct));
         BNNodeSetUserData(bnn, data);

         data->avgDataLL = 0;

         data->isChangedFromCurrent = 0;
         data->p0 = 1;

         data->current = 0;
      }

      /* flag to keep it from getting inited too many times */
      BNSetUserData(current, (void *)1);
   }
}

static BeliefNetNode _BNGetNodeByID(BNUserData netData, int id) {
   if(netData->changedOne) {
      if(BNNodeGetID(netData->changedOne) == id) {
         return netData->changedOne;
      }
   }

   return BNGetNodeByID(netData->currentNet, id);
}

static double _CalculateCP(BeliefNetNode bnn, double event, double row) {
   /* HERE HERE Heuristic hack */
   return (event + 1.0) / (row + (float)BNNodeGetNumValues(bnn));
   //return event / row;
}

static double _GetEpsilonNormal(BNUserData firstData, BNUserData secondData,
                                                       BeliefNet current) {
   BeliefNetNode bnn;
   int i, j, k, numCPTRows, n1;
   double likelihood, numSamples;
   double bound;
   StatTracker st;
   
   bound = 0;

   //n1 = -1;
   if(firstData->changedOne) {
      bnn = firstData->changedOne;
      n1 = BNNodeGetID(bnn);
      st = StatTrackerNew();

      numCPTRows = BNNodeGetNumCPTRows(bnn);
      numSamples = BNNodeGetNumSamples(bnn);
      for(j = 0 ; j < numCPTRows ; j++) {
         /* HACK for efficiency break BNN ADT */
         for(k = 0 ; k < BNNodeGetNumValues(bnn) ; k++) {
            likelihood = (bnn->eventCounts[j][k] / numSamples) * 
               log(_CalculateCP(bnn, bnn->eventCounts[j][k],
                                     bnn->rowCounts[j]));
            for(i = 0 ; i < bnn->eventCounts[j][k] ; i++) {
               StatTrackerAddSample(st, likelihood);
            }
         }
      }

      bound += StatTrackerGetNormalBound(st, gDelta);
      StatTrackerFree(st);

      bnn = _BNGetNodeByID(secondData, n1);
      st = StatTrackerNew();

      numCPTRows = BNNodeGetNumCPTRows(bnn);
      numSamples = BNNodeGetNumSamples(bnn);
      for(j = 0 ; j < numCPTRows ; j++) {
         /* HACK for efficiency break BNN ADT */
         for(k = 0 ; k < BNNodeGetNumValues(bnn) ; k++) {
            likelihood = (bnn->eventCounts[j][k] / numSamples) * 
               log(_CalculateCP(bnn, bnn->eventCounts[j][k],
                                     bnn->rowCounts[j]));
            for(i = 0 ; i < bnn->eventCounts[j][k] ; i++) {
               StatTrackerAddSample(st, likelihood);
            }
         }
      }

      bound += StatTrackerGetNormalBound(st, gDelta);
      StatTrackerFree(st);
   } else if(secondData->changedOne) {
      bnn = secondData->changedOne;
      n1 = BNNodeGetID(bnn);
      st = StatTrackerNew();

      numCPTRows = BNNodeGetNumCPTRows(bnn);
      numSamples = BNNodeGetNumSamples(bnn);
      for(j = 0 ; j < numCPTRows ; j++) {
         /* HACK for efficiency break BNN ADT */
         for(k = 0 ; k < BNNodeGetNumValues(bnn) ; k++) {
            likelihood = (bnn->eventCounts[j][k] / numSamples) * 
               log(_CalculateCP(bnn, bnn->eventCounts[j][k],
                                     bnn->rowCounts[j]));
            for(i = 0 ; i < bnn->eventCounts[j][k] ; i++) {
               StatTrackerAddSample(st, likelihood);
            }
         }
      }

      bound += StatTrackerGetNormalBound(st, gDelta);
      StatTrackerFree(st);

      bnn = _BNGetNodeByID(firstData, n1);
      st = StatTrackerNew();

      numCPTRows = BNNodeGetNumCPTRows(bnn);
      numSamples = BNNodeGetNumSamples(bnn);
      for(j = 0 ; j < numCPTRows ; j++) {
         /* HACK for efficiency break BNN ADT */
         for(k = 0 ; k < BNNodeGetNumValues(bnn) ; k++) {
            likelihood = (bnn->eventCounts[j][k] / numSamples) * 
               log(_CalculateCP(bnn, bnn->eventCounts[j][k],
                                     bnn->rowCounts[j]));
            for(i = 0 ; i < bnn->eventCounts[j][k] ; i++) {
               StatTrackerAddSample(st, likelihood);
            }
         }
      }

      bound += StatTrackerGetNormalBound(st, gDelta);
      StatTrackerFree(st);
   }

   return bound;
}


static double _GetNodeScoreRange(BeliefNetNode bnn) {
   double p0;

   BNNodeUserData data = BNNodeGetUserData(bnn);

   if(data == 0) {
      p0 = (1.0 / (5.0 * (double)BNNodeGetNumValues(bnn)));
   } else {
      p0 = min(1.0 / (5.0 * (double)BNNodeGetNumValues(bnn)), 
               data->p0 / (double)gP0Multiplier);
   }

   return fabs(log(p0));
}

static double _GetComparedNodesScoreRange(BNUserData firstData,
                                              BNUserData secondData) {
   double scoreRange;

  
   scoreRange = 0;

   if(firstData->changedOne) {
      scoreRange += _GetNodeScoreRange(firstData->changedOne);
   } else if(secondData->changedOne) {
      scoreRange += _GetNodeScoreRange(secondData->changedOne);
   }

   return scoreRange;
}

void _UpdateNodeAveDataLL(BeliefNetNode bnn) {
   int numCPTRows;
   BNNodeUserData data;
   int j, k;
   double numSamples;
   double prob, logCP;

   data = BNNodeGetUserData(bnn);

   /* if this is the current node or is changed from it */
   if(data->isChangedFromCurrent || (data->current == 0)) {
      data->avgDataLL = 0;
      numCPTRows = BNNodeGetNumCPTRows(bnn);
      numSamples = BNNodeGetNumSamples(bnn);

      for(j = 0 ; j < numCPTRows ; j++) {
         /* HACK for efficiency break BNN ADT */
         for(k = 0 ; k < BNNodeGetNumValues(bnn) ; k++) {
            if(bnn->eventCounts[j][k] > 0) {
               prob = bnn->eventCounts[j][k] / numSamples;

               /* HERE HERE Heuristic Hack */
               //logCP = log(bnn->eventCounts[j][k] / bnn->rowCounts[j]);
               logCP = log(_CalculateCP(bnn, 
                            bnn->eventCounts[j][k], bnn->rowCounts[j]));

               if(data->p0 > bnn->eventCounts[j][k] / bnn->rowCounts[j] && 
                           bnn->eventCounts[j][k] / bnn->rowCounts[j] > 0) {
                  data->p0 = bnn->eventCounts[j][k] / bnn->rowCounts[j];
               }

               data->avgDataLL += prob * logCP;
            }
         }
      }
   }
}

static void _UpdateNetScore(BNUserData netData) {
   if(netData->changedOne) {
      _UpdateNodeAveDataLL(netData->changedOne);
   }
   //if(netData->changedTwo) {
   //   _UpdateNodeAveDataLL(netData->changedTwo);
   //}
}

static double _GetNodeScore(BeliefNetNode bnn) {
   BNNodeUserData data = BNNodeGetUserData(bnn);
   
   if(data->current == 0 || data->isChangedFromCurrent) {
      return data->avgDataLL;
   } else {
      return data->current->avgDataLL;
   }
}

double _GetDeltaScore(BNUserData firstData, BNUserData secondData) {
   double scoreOne, scoreTwo;
   
   scoreOne = scoreTwo = 0;

   if(firstData->changedOne) {
      scoreOne += _GetNodeScore(firstData->changedOne);
      scoreTwo += _GetNodeScore(_BNGetNodeByID(secondData, 
                                BNNodeGetID(firstData->changedOne)));
   } else if(secondData->changedOne) {
      scoreOne += _GetNodeScore(_BNGetNodeByID(firstData, 
                                BNNodeGetID(secondData->changedOne)));
      scoreTwo += _GetNodeScore(secondData->changedOne);
   }

   DebugMessage(1, 4, "    first %lf second %lf\n", scoreOne, scoreTwo);
   return scoreOne - scoreTwo;
}

//static int _IsFirstNetBetter(BNUserData firstData, BNUserData secondData) {
//   return _GetDeltaScore(firstData, secondData) >= 0;
//}
#define _IsFirstNetBetter(firstData,secondData) ((int)(_GetDeltaScore(firstData,secondData) >= 0))

//static int _IsCurrentNet(BNUserData data) {
//   return (data->changedOne == 0);
//             // && (data->changedTwo == 0);
//}
#define _IsCurrentNet(data) ((int)(((BNUserData)(data))->changedOne == 0))

static void _PickWinnerInTieFreeRest(ModelSearch ms, BNUserData winner) {
   int i;

   for(i = VLLength(ms->choices) - 1 ; i >= 0 ; i--) {
      //if(_IsCurrentNet(VLIndex(ms->choices, i))) {
      //   /* This favors the current net in ties */
      //   gNumCurrentInTie++;
      //   _FreeUserData(winner);
      //   BNFree(winner);
      //   winner = VLRemove(ms->choices, i);
      //} else {
         BNUserData net = VLIndex(ms->choices, i);
         if (net != winner) {
            _FreeUserData(net);
            // OPTIMIZE: Free and add winner, don't remove N times
            VLRemove(ms->choices, i);
         }
      //}
   }

   if(_IsCurrentNet(winner)) {
      gNumCurrentInTie++;
   }
   FLUSH;
}

void _GetOneStepChoicesForSearch(ModelSearch ms) {
   int k;
   BeliefNetNode currentNode, parentNode;
   //BNNodeUserData data;
   BNUserData netData;
   long preAllocation;

   ms->initialExampleNumber = -1;
   ms->initialExamplePass = 0;

   DebugMessage(VLLength(ms->choices) > 0, 0,
            "******Warning list in choices not empty\n");

   preAllocation = MGetTotalAllocation();

   /* the no-change-net */
   VLAppend(ms->choices,
                 _InitUserData(ms->nodeID, 0, -1, 0, ms->currentModel));

   DebugMessage(1, 3, " allocated current net for node %d, size %ld\n", ms->nodeID,
        MGetTotalAllocation() - preAllocation);

   for(k = 0 ; (ms->allowedParents && k<VLLength(ms->allowedParents)) ||
               (!ms->allowedParents && k < BNGetNumNodes(ms->currentModel)) ;  k++) {
      
      int iNewParent = (ms->allowedParents? (int)VLIndex(ms->allowedParents, k) : k);
      int linkChangeIndex = k;
      if(iNewParent != ms->nodeID) {
         currentNode = BNGetNodeByID(ms->currentModel, ms->nodeID);
         parentNode = BNGetNodeByID(ms->currentModel, iNewParent);

         /* if they aren't related consider adding dst as a parent */
         if(BNNodeLookupParentIndex(currentNode, parentNode) == -1 &&
               BNNodeLookupParentIndex(parentNode, currentNode) == -1) {

            preAllocation = MGetTotalAllocation();

            netData  = _InitUserData(ms->nodeID, 1, iNewParent, 1, ms->currentModel);

            DebugMessage(1, 4, " allocated new net, size %ld\n",
                   MGetTotalAllocation() - preAllocation);

            if((gMaxParentsPerNode == -1 ||
                BNNodeGetNumParents(netData->changedOne) <= 
                                            gMaxParentsPerNode) &&

               (gMaxParameterCount == -1 ||
                 (BNNodeGetNumParameters(netData->changedOne) <=
                                                gMaxParameterCount)) &&

               (ms->linkChangeCounts[linkChangeIndex] <= 2) &&
              (gLimitBytes == -1 ||
            ((MGetTotalAllocation() - preAllocation) <=  gMaxBytesPerModel))&&
                   (gLazyCycleCheck || !_BNHasCycle(netData))) {
               
               VLAppend(ms->choices, netData);
            } else {
               if(gMaxParentsPerNode != -1 &&
                        BNNodeGetNumParents(netData->changedOne) > 
                                                    gMaxParentsPerNode) {
                  gNumByParentLimit++;
               } else if(gLimitBytes != -1 &&
                       ((MGetTotalAllocation() - preAllocation) >
                                                gMaxBytesPerModel)) {
                  gNumByMemoryLimit++;
               } else if(ms->linkChangeCounts[linkChangeIndex] > 2) {
                  gNumByChangeLimit++;
               } else if(gMaxParameterCount != -1 && 
                     (BNNodeGetNumParameters(netData->changedOne)
                                              > gMaxParameterCount)) {
                  gNumByParameterLimit++;
               }

               _FreeUserData(netData);
            }
         } else if(gAllowRemove && 
                 BNNodeLookupParentIndex(currentNode, parentNode) != -1) {
            /* Consider removing the link */
            preAllocation = MGetTotalAllocation();

            netData = _InitUserData(ms->nodeID, 1, iNewParent, 0, ms->currentModel);

            DebugMessage(1, 3, " allocated new net, size %ld\n",
                   MGetTotalAllocation() - preAllocation);

            if((gMaxParameterCount == -1 ||
                    (BNNodeGetNumParameters(netData->changedOne) <= 
                                                  gMaxParameterCount)) &&

               (ms->linkChangeCounts[linkChangeIndex] <= 2) &&

               (gLimitBytes == -1 ||
                      ((MGetTotalAllocation() - preAllocation) 
                                   <=  gMaxBytesPerModel)) &&
                  (gLazyCycleCheck || !_BNHasCycle(netData))) {
               VLAppend(ms->choices, netData);
            } else {
               if(gLimitBytes != -1 &&
                       ((MGetTotalAllocation() - preAllocation) >
                                                gMaxBytesPerModel)) {
                  gNumByMemoryLimit++;
               } else if(ms->linkChangeCounts[linkChangeIndex] > 2) {
                  gNumByChangeLimit++;
               } else if(gMaxParameterCount != -1 && 
                     (BNNodeGetNumParameters(netData->changedOne) > 
                                                     gMaxParameterCount)) {
                  gNumByParameterLimit++;
               }

               _FreeUserData(netData);
            }
         }
      }
   }
}

// Returns NULL for _secondBest if there is only one model
void UpdateScoresAndFindBestWithoutCheckingCycles(
   ModelSearch ms, BNUserData *_best, BNUserData *_secondBest) {
   BNUserData bestNet=NULL, secondBestNet=NULL, compareNet=NULL;
   int i;

   bestNet = VLIndex(ms->choices, 0);
   _UpdateNetScore(bestNet);

   if (VLLength(ms->choices) > 1) {
      compareNet = VLIndex(ms->choices, 1);
      _UpdateNetScore(compareNet);
      if (!_IsFirstNetBetter(bestNet, compareNet)) {
         secondBestNet = bestNet;
         bestNet = compareNet;
      }
      else
         secondBestNet = compareNet;

      for(i = 2 ; i < VLLength(ms->choices) ; i++) {
         compareNet = VLIndex(ms->choices, i);
         _UpdateNetScore(compareNet);

         if(!_IsFirstNetBetter(bestNet, compareNet)) {
            secondBestNet = bestNet;
            bestNet = compareNet;
         } else if (!_IsFirstNetBetter(secondBestNet, compareNet)) {
            secondBestNet = compareNet;
         }
      }
   }
   *_best=bestNet;
   *_secondBest = secondBestNet;
   FLUSH;
}


// Returns NULL for _secondBest if there is only one model
void UpdateScoresAndFindBestWithCheckingCycles(
   ModelSearch ms, BNUserData *_best, BNUserData *_secondBest) {
   BNUserData bestNet=NULL, secondBestNet=NULL, compareNet=NULL;
   int i;

   for(i=VLLength(ms->choices)-1; i>=0; i--) {
     compareNet = VLIndex(ms->choices, i);
     _UpdateNetScore(compareNet);
     if (!bestNet || !_IsFirstNetBetter(bestNet, compareNet)) {
        // Potential best. Check for cycles
        if (_BNHasCycle(compareNet)) {
           _FreeUserData(compareNet);
           VLRemove(ms->choices, i);
        }
        else {
           secondBestNet = bestNet;
           bestNet = compareNet;
        }
     }
     else if (!secondBestNet || !_IsFirstNetBetter(secondBestNet, compareNet)) {
        // Potential second-best. Check for cycles
        if (_BNHasCycle(compareNet)) {
           _FreeUserData(compareNet);
           VLRemove(ms->choices, i);
        }
        else
           secondBestNet = compareNet;
     }
   }
   *_best=bestNet;
   *_secondBest = secondBestNet;
   FLUSH;
}

/*
// OPTIMIZE: Make a macro
int heap_priority(void *notused, BNUserData net1, BNUserData net2) {
   // Should return 1,0,-1
   //  -1 if net2 is better than net1
   if (_GetDeltaScore(net1, net2) == 0 && net1 != net2)
      printf("EQUAL NETS FOUND\n");
   return (_IsFirstNetBetter(net1, net2)? 1:-1);
}

// Returns NULL for _secondBest if there is only one model
// OPTIMIZE: allow passing _secondBest as null to mean that you don't 
//  need to find the second best
void UpdateScoresAndFindBestWithCheckingCycles(
   ModelSearch ms, BNUserData *_best, BNUserData *_secondBest) {
   BNUserData bestNet=NULL, secondBestNet=NULL, compareNet=NULL;
   int i, foundCycle;

   // Priority Queue data pieces
   int heap_max = VLLength(ms->choices);
   int heap_size = 0;
   BNUserData heap_tmp;
   BNUserData *heap_array = MNewPtr(heap_max * sizeof(BNUserData));

   // First, put all the nets in to the priority queue
   for(i=VLLength(ms->choices)-1; i>=0; i--) {
      compareNet = VLIndex(ms->choices, i);
      _UpdateNetScore(compareNet);
      compareNet->myTmpInternalData=0;
      WBHEAP_ADD(heap_array, heap_size, heap_max, heap_priority, NULL, compareNet);
   }
   // Now, find the best two which have no cycles
   foundCycle=0;
   while(WBHEAP_VALIDITEM(heap_size,0)) {
      compareNet = WBHEAP_GETITEM(heap_array, 0);
      WBHEAP_DELETE(heap_array, heap_size, heap_priority, NULL, heap_tmp, 0);
      if (_BNHasCycle(compareNet)) {
         compareNet->myTmpInternalData=1; // remove this one later
         foundCycle=1;
      }
      else {
         // Good network. Put in best or second best, depending
         if (!bestNet)
            bestNet = compareNet;
         else {
            secondBestNet = compareNet;
            assert(_IsFirstNetBetter(bestNet, secondBestNet));
            break;
         }
      }
   }

//   if (foundCycle) {
//      // Remove all the entries marked with tmpInternalData=1
//      int numChoices = VLLength(ms->choices);
//      int currentIndex=0;
//      for (i=0; i<numChoices; i++) {
//         BNUserData neti = VLIndex(ms->choices, i);
//         if (neti->myTmpInternalData == 0) 
//            VLSet(ms->choices, currentIndex++, neti);
//         else
//            _FreeUserData(neti);
//      }
//      VLTruncate(ms->choices, currentIndex);
//   }

   if (foundCycle) {
      for (i=VLLength(ms->choices)-1; i>=0; i--) {
         BNUserData net = VLIndex(ms->choices, i);
         if (net->myTmpInternalData == 1) {
            _FreeUserData(net);
            VLRemove(ms->choices, i);
         }
      }
   }
   

      
   *_best=bestNet;
   *_secondBest = secondBestNet;
   MFreePtr(heap_array);
   FLUSH;
}

*/

// May return with no winners if they all had cycles
static void _PickCurrentBestAsWinner(ModelSearch ms) {
   int i;
   BNUserData bestNet, secondNet;

   _UpdateNodeAveDataLL(BNGetNodeByID(ms->currentModel, ms->nodeID));

   if (!gLazyCycleCheck) 
      UpdateScoresAndFindBestWithoutCheckingCycles(ms, &bestNet, &secondNet);
   else 
      UpdateScoresAndFindBestWithCheckingCycles(ms, &bestNet, &secondNet);
   
   if (!bestNet) {
      //There are no nets left. Just return
      return;
   }

   // Free everyone but the best
   for(i = VLLength(ms->choices) - 1 ; i >= 0 ; i--) {
      BNUserData net = VLIndex(ms->choices, i);
      if (net != bestNet)
         _FreeUserData(net);
   }
   VLTruncate(ms->choices, 0);
   VLAppend(ms->choices, bestNet);
   if(_IsCurrentNet(bestNet)) 
      gNumCurrentInTie++;
}

static void _CompareNetsBoundFreeLoosers(ModelSearch ms) {
//BeliefNet current, VoidListPtr netChoices) {

   int i;
   BNUserData bestNet, secondNet, compareNet;
   double deltaScore;
   double epsilon;
   double scoreRangeNode;

   _UpdateNodeAveDataLL(BNGetNodeByID(ms->currentModel, ms->nodeID));

   // OPTIMIZE: Could actually ignore cycles here too, if it speeds things
   //  up a bunch, it is approximately correct to ignore cycles
   if (!gLazyCycleCheck) 
      UpdateScoresAndFindBestWithoutCheckingCycles(ms, &bestNet, &secondNet);
   else 
      UpdateScoresAndFindBestWithCheckingCycles(ms, &bestNet, &secondNet);

   if (!secondNet) {
      // There are only zero or one nets
      // ERROR: I don't know if this is the right thing to do here
      //   for when there is just one net. Should it automatically win?
      return;
   }

   /* see if the bestNet beats the secondNet */
   deltaScore = _GetDeltaScore(bestNet, secondNet);

   scoreRangeNode = fabs(_GetComparedNodesScoreRange(bestNet, secondNet));

   /* HERE for missing data or unequal in compare */
   epsilon = _GetEpsilonNormal(bestNet, secondNet, ms->currentModel);

   DebugMessage(1, 3, " delta %lf epsilon %lf nodeRange: %f times-tau: %f\n",
          deltaScore, epsilon, scoreRangeNode, scoreRangeNode * gTau);
   if(deltaScore - epsilon > 0) {
      DebugMessage(1, 2,
               "Finish by winner, diff %f, epsilon %f were %d left\n", 
               deltaScore, epsilon, VLLength(ms->choices));

      gNumByWin += VLLength(ms->choices) - 1;
      gNumBoundsUsed += VLLength(ms->choices) - 1;

      if(_IsCurrentNet(bestNet)) {
         gNumCurrentByWin++;
      }

      for(i = VLLength(ms->choices) - 1 ; i >= 0 ; i--) {
         BNUserData net = VLIndex(ms->choices, i);
         if (net != bestNet) {
            _FreeUserData(net);
            // OPTIMIZE: Clear and add, rather than N removes
            VLRemove(ms->choices, i);
         }
      }
   } else {
      /* tie if it won't cost more than range * tau */
      if(fabs(deltaScore) <= epsilon &&
                    epsilon <= (gTau * scoreRangeNode)) {
         /* that's a tie */

         DebugMessage(1, 2, 
                "Finish by tie diff %f epsilon %f range %f with %d left\n",
             deltaScore, epsilon, scoreRangeNode, VLLength(ms->choices));
         //DebugMessage(1, 3, "  best %p compare %p\n", bestNet, secondNet);

         gNumByTie += VLLength(ms->choices) - 1;
         gNumBoundsUsed += VLLength(ms->choices) - 1;

         _PickWinnerInTieFreeRest(ms, bestNet);

      } else if(fabs(deltaScore) < epsilon) {
         /* see if we could possibly call a tie */
         //if(tauNeededToQuitNode < epsilon / scoreRangeNode) {
         //   tauNeededToQuitNode = epsilon / scoreRangeNode;
         //}
      }
   }

   /* if not do another pass deactivating clear loosers */
   if(VLLength(ms->choices) > 1) {
      for(i = VLLength(ms->choices) - 1 ; i >= 0 ; i--) {
         compareNet = VLIndex(ms->choices, i);
         if (compareNet != bestNet && compareNet != secondNet) {
            deltaScore = _GetDeltaScore(secondNet, compareNet);
            scoreRangeNode = 
                  fabs(_GetComparedNodesScoreRange(secondNet, compareNet));

            /* HERE for missing data or unequal in compare */
            epsilon = _GetEpsilonNormal(secondNet, compareNet,
                                                ms->currentModel);

            if(deltaScore - epsilon > 0) {
               DebugMessage(1, 2, 
                 "Early freeing a net, diff %f, epsilon %f now %d left\n", 
                  deltaScore, epsilon, VLLength(ms->choices));
               gNumByWin++;
               gNumBoundsUsed++;

               if(_IsCurrentNet(compareNet)) {
                  DebugMessage(1, 2, 
                    " Current best freed bound - delta %lf epsilon %lf.\n", 
                                                    deltaScore, epsilon);
               }
               _FreeUserData(VLRemove(ms->choices, i));
            } 
         }
      }
   }
}

// allowedParents may be NULL, which means allow any node to be a parent
//  otherwise, it's a list of integers
ModelSearch MSNew(BeliefNet currentBN, int nodeID, VoidListPtr allowedParents) {
   int i, possibleParents;
   ModelSearch ms = MNewPtr(sizeof(ModelSearchStruct));

   ms->initialExampleNumber = -1;
   ms->initialExamplePass = 0;
   ms->lastParentAdded = -1;
   ms->isActive = 0;

   ms->choices = VLNew();
   ms->isDone = 0;
   ms->doneCollectingExamples = 0;

   ms->currentModel = currentBN;
   ms->nodeID       = nodeID;
   ms->allowedParents = allowedParents;

   /* MEM this may leak some memory, but should only happen once */
   _InitCurrentUserData(currentBN);

   // If !allowedParents, then linkChangeCounts[i] is for node i
   // if allowedParents, then linkChangeCounts[i] is for node allowedParents[i]
   //   note, if you are starting with a network which includes parents that are
   //    not in your parentset, those parents are non-removable (see also GetOneStepChoicesForSearch)
   possibleParents = (ms->allowedParents? VLLength(ms->allowedParents): BNGetNumNodes(currentBN));
   ms->linkChangeCounts = MNewPtr(sizeof(int) * possibleParents);
   for(i = 0 ; i < possibleParents ; i++) {
      ms->linkChangeCounts[i] = 0;
   }

   return ms;
}

void MSFree(ModelSearch ms) {
   MSDeactivate(ms);
   VLFree(ms->choices);
   MFreePtr(ms->linkChangeCounts);
   MFreePtr(ms);
}

int MSGetNodeID(ModelSearch ms) {
   return ms->nodeID;
}


/*

void MSAddExample(ModelSearch ms, ExamplePtr e, long exampleNumber) {
   int i;
   BNUserData netData;

   if(ms->initialExampleNumber == -1) {
      ms->initialExampleNumber = exampleNumber;
   } else if(exampleNumber == ms->initialExampleNumber) {
      // we are at the end of looping through all the examples 
      //  so just pick the best one as the winner and maybe print warning 

      ms->doneCollectingExamples = 1;

      gNumUsedWholeDB++;

      DebugMessage(1, 1, "******Looped through all examples\n");
   }

   if(!ms->doneCollectingExamples) {
      BNNodeAddSample(BNGetNodeByID(ms->currentModel, ms->nodeID), e);
      for(i = 0 ; i < VLLength(ms->choices) ; i++) {
         netData = VLIndex(ms->choices, i);
         if(netData->changedOne) {
            BNNodeAddSample(netData->changedOne, e);
         }
      }
   }
}
*/

// initialNum and initialNumPassNumber are for when this MS was first given samples
//   num is the example number. passnumber is 0,1,.. the number of times we have
//   gone through the entire set of data.

// If finalPassNumber - initialPassNumber > 1 then we've seen it all
// If finalPassNumber - initialPassNumber = 1 then 
//     If finalNum >= initialNum, then we've seen it all
//     else we haven't
// If finalPassNumber - initialPassNumber = 0 then we haven't
#define _HaveSeenAllExamples(initialNum, initialPassNumber, finalNum, finalPassNumber) \
   ( (((finalPassNumber) - (initialPassNumber)) > 1) || \
    ((((finalPassNumber) - (initialPassNumber)) == 1) && ((finalNum) >= (initialNum))) \
   )

void MSAddExamples(ModelSearch ms, VoidListPtr exs, long finalExampleNumber, int finalExamplePass) {
   int i;
   BNUserData netData;

   if(ms->initialExampleNumber == -1) {
      ms->initialExampleNumber = finalExampleNumber;
      ms->initialExamplePass = finalExamplePass;
   }

   if ( _HaveSeenAllExamples(ms->initialExampleNumber, ms->initialExamplePass, finalExampleNumber, finalExamplePass)) {
      /* we are at the end of looping through all the examples */
      /*  so just pick the best one as the winner and maybe print warning */
      ms->doneCollectingExamples = 1;
      gNumUsedWholeDB++;
      DebugMessage(1, 2, "******Looped through all examples\n");
   }

   if(!ms->doneCollectingExamples) {
      BNNodeAddSamples(BNGetNodeByID(ms->currentModel, ms->nodeID), exs);
      for(i = 0 ; i < VLLength(ms->choices) ; i++) {
         netData = VLIndex(ms->choices, i);
         if(netData->changedOne) {
            BNNodeAddSamples(netData->changedOne, exs);
         }
      }
   }

}


void MSTerminateSearch(ModelSearch ms) {
   ms->doneCollectingExamples = 1;
}

int MSIsDone(ModelSearch ms) {
   return ms->isDone;
}

// Assumes that winner has been checked to make sure it has no cycles
//  regardless of lazy flag
static void _ApplyWinner(ModelSearch ms) {
   BNUserData winnerData;
   BNNodeUserData nodeData;
   BeliefNetNode changedNode;
   long samples;

   winnerData = VLRemove(ms->choices, 0);

   samples = BNNodeGetNumSamples(_BNGetNodeByID(winnerData, ms->nodeID));

   if(samples > 0) {
      gMinSamplesInDecision = min(samples, gMinSamplesInDecision);
      gMaxSamplesInDecision = max(samples, gMaxSamplesInDecision);
      gTotalSamplesInDecision += samples;
   } else {
      gNumZeroSamplesInDecision++;
   }


   /* if doing nothing wins then set isDone */
   if(_IsCurrentNet(winnerData)) {
      DebugMessage(1, 2, "Current net wins for %d\n", ms->nodeID);
      ms->isDone = 1;
   } else if(!_BNHasCycle(winnerData)) {
      if (!ms->allowedParents)
         ms->linkChangeCounts[winnerData->changedParentID]++;
      else {
         assert(VLFind(ms->allowedParents, (void*)winnerData->changedParentID) != -1);
         ms->linkChangeCounts[VLFind(ms->allowedParents, (void*)winnerData->changedParentID)]++;
      }         

      /* apply any winner to the current model */
      BNFlushStructureCache(ms->currentModel);
      BNNodeFreeCPT(BNGetNodeByID(ms->currentModel, ms->nodeID));

      if(winnerData->changeComplexity == -1) {
         gNumRemoved++;

         changedNode = BNGetNodeByID(ms->currentModel, ms->nodeID);
         BNNodeRemoveParent(changedNode,
           BNNodeLookupParentIndexByID(changedNode,
                                 winnerData->changedParentID));

         nodeData = BNNodeGetUserData(changedNode);
         nodeData->p0 = 1.0;

         DebugMessage(1, 2, "Winner for %d remove parent %d now %d parents\n", 
            ms->nodeID, winnerData->changedParentID,
             BNNodeGetNumParents(BNGetNodeByID(ms->currentModel, ms->nodeID)));
      } else if(winnerData->changeComplexity == 2) {
         gNumAdded++;

         changedNode = BNGetNodeByID(ms->currentModel, ms->nodeID);
         BNNodeAddParent(changedNode,
              BNGetNodeByID(ms->currentModel, winnerData->changedParentID));

         nodeData = BNNodeGetUserData(changedNode);
         nodeData->p0 = 1.0;

         DebugMessage(1, 2, "Winner for %d new parent %d now %d parents\n", 
            ms->nodeID, winnerData->changedParentID,
             BNNodeGetNumParents(BNGetNodeByID(ms->currentModel, ms->nodeID)));
      }


      BNNodeInitCPT(BNGetNodeByID(ms->currentModel, ms->nodeID));

      /* EFF I don't think I need this SetCPTFrom call */
      //BNNodeSetCPTFrom(BNGetNodeByID(ms->currentModel, ms->nodeID),
      //              BNGetNodeByID(winnerData->changedOne, ms->nodeID));

   }

   _FreeUserData(winnerData);

   MSDeactivate(ms);
}

int MSCheckForWinner(ModelSearch ms) {
   int foundWinner;

   if(MSIsDone(ms) || !MSIsActive(ms)) {
      foundWinner = 0;
   } else if(ms->doneCollectingExamples) {
      _PickCurrentBestAsWinner(ms);
      if (VLLength(ms->choices) == 1)
         foundWinner = 1;
      else {
         foundWinner = 0;
         MSDeactivate(ms);
      }
   } else {
      _CompareNetsBoundFreeLoosers(ms);

      if(VLLength(ms->choices) == 1) {
         foundWinner = 1;
      } else { 
         foundWinner = 0;
         if (VLLength(ms->choices) == 0) 
            MSDeactivate(ms);
      }
   }

   if(foundWinner) {
      _ApplyWinner(ms);
   }

   return foundWinner;
}

void MSModelChangedHandleConflicts(ModelSearch ms) {
   int i;
   BNUserData netData;

   /* apply the changes to each of the search alternatives */
   /* check each for cycles and if there are any free them */
   if (gLazyCycleCheck) {
      if (VLLength(ms->choices)==1) {
         netData = VLIndex(ms->choices, 0);
         if (_BNHasCycle(netData)) {
            gNumByCycleConflict++;
            _FreeUserData(netData);
            VLRemove(ms->choices, 0);
         }
         else {
            if (_IsCurrentNet(netData))
               gNumCurrentByDefault++;
            _ApplyWinner(ms);
         }
      }
      if (VLLength(ms->choices) == 0) {
         MSDeactivate(ms);
      }

   }
   else {
      for(i = VLLength(ms->choices) - 1 ; i >= 0 ; i--) {
         netData = VLIndex(ms->choices, i);
         if(_BNHasCycle(netData)) {
            gNumByCycleConflict++;
            _FreeUserData(netData);
            VLRemove(ms->choices, i);

         //} else if(gMaxParameterCount != -1 &&
         //              (BNGetNumParameters(bn) > gMaxParameterCount)) {
         //      gNumByParameterConflict++;
         //      _FreeUserData(netData);
         //      VLRemove(ms->choices, i);
         }
      }
      if(VLLength(ms->choices) == 1) {
         /* if only one left then handle the winner */
         if(_IsCurrentNet(VLIndex(ms->choices, 0))) {
            gNumCurrentByDefault++;
         }
         _ApplyWinner(ms);
      } else if(VLLength(ms->choices) == 0) {
         /* if everyone is free then deactivate and let someone else cope */   
         MSDeactivate(ms);
      }
   }


}


int MSIsActive(ModelSearch ms) {
   return ms->isActive;
}

void MSActivate(ModelSearch ms) {
   if(!ms->isActive) {
      _GetOneStepChoicesForSearch(ms);
      ms->isActive = 1;
      ms->doneCollectingExamples = 0;

      ms->numSearchOptions = VLLength(ms->choices);

      DebugMessage(1, 3, "    activated with %d alternatives\n",
                              ms->numSearchOptions);

      if(ms->numSearchOptions <= 1) {
         gNumCurrentByDefault++;
      }
   }
}

void MSDeactivate(ModelSearch ms) {
   int i;

   if(ms->isActive) {
      for(i = VLLength(ms->choices) - 1 ; i >= 0 ; i--) {
         _FreeUserData(VLRemove(ms->choices, i));
      }

      ms->isActive = 0;
   }
}

int MSNumActiveAlternatives(ModelSearch ms) {
   return VLLength(ms->choices);
}
int MSDoneCollectingExamples(ModelSearch ms) {
   return ms->doneCollectingExamples;
}
