
/***************************************************************************\
**                                                                         **
** WW     WW IIIIIIII NNN   NN DDDDDDD  BBBBBBB     AA     SSSSSS EEEEEEEE **
** WW  W  WW    II    NNNN  NN DD    DD BB    BB  AA  AA  SS      EE       **
** WW  W  WW    II    NN NN NN DD    DD BBBBBBB  AAAAAAAA  SSSSSS EEEEEE   **
**  WW W WW     II    NN  NNNN DD    DD BB    BB AA    AA      SS EE       **
**   WWWWW   IIIIIIII NN   NNN DDDDDDD  BBBBBBB  AA    AA SSSSSS  EEEEEEEE **
**                                                                         **
**  SSSSSS  OOOOOO  FFFFFFFF TTTTTTTT WW     WW    AA    RRRRRRR  EEEEEEEE **
** SS      OO    OO FF          TT    WW  W  WW  AA  AA  RR    RR EE       **
**  SSSSS  OO    OO FFFFF       TT    WW  W  WW AAAAAAAA RRRRRRR  EEEEEE   **
**      SS OO    OO FF          TT     WW W WW  AA    AA RR   RR  EE       **
** SSSSSS   OOOOOO  FF          TT      WWWWW   AA    AA RR    RR EEEEEEEE **
**                                                                         **
********** NOTICE ***********************************************************
**       This file contains valuable trade secrets and proprietary         **
**       assets of Windbase Software Inc.  Embodying substantial           **
**       creative efforts and confidential information.  Unauthorized      **
**       use, copying, decompiling, translating, disclosure or             **
**       transfer, of any kind, is strictly prohibited.                    **
**                                                                         **
**       COPYRIGHT (C) 1992 - 1996.  Windbase Software Inc.                **
**       ALL RIGHTS RESERVED.                                              **
\***************************************************************************/
//Provided only so we may continue to use our source.  You may use these
//macros freely.

/*************************************************************************\
**  Priority Heap Structure and Prototypes
**
** NOTE: The macros are defined below.
**
** While these macros allow you to work with your arrays directly, the
** MemSL provides a friendlier user interface to the priority heap
** functionality.  The functions provided with the MemSL are:
**
**          WBPHEAP *WBPHeapOpen(WBERROR *, unsigned int);
**          void WBPHeapClose(WBPHEAP *);
**          #ifdef WBTHREADS
**            void WBPHeapLockF(WBPHEAP *, void *, int (*)(void *));
**            void WBPHeapUnlockF(WBPHEAP *, void *, int (*)(void *));
**            void WBPHeapRLockF(WBPHEAP *, void *, int (*)(void *));
**            void WBPHeapRUnlockF(WBPHEAP *, void *, int (*)(void *));
**          #endif
**          void WBPHeapPriorityF(WBPHEAP *, void *,
**                                int (*)(void *, void *, void *));
**          void WBPHeapCompareF(WBPHEAP *, void *,
**                               int (*)(void *, void *, void *));
**          void WBPHeapDeleteF(WBPHEAP *, void *,
**                              void (*)(void *, void *));
**          void WBPHeapExecuteF(WBPHEAP *, void *,
**                               void (*)(void *, void *));
**          int WBPHeapAdd(WBPHEAP *, void *);
**          int WBPHeapDelete(WBPHEAP *, void *);
**          void WBPHeapClear(WBPHEAP *);
**          void *WBPHeapSearch(WBPHEAP *, void *);
**          void *WBPHeapGet(WBPHEAP *);
**          void *WBPHeapPeek(WBPHEAP *);
**          void WBPHeapExecute(WBPHEAP *);
**          long WBPHeapNumItems(WBPHEAP *);
**          int WBPHeapIsEmpty(WBPHEAP *);
**
** The above style and functionality is used throughout the MemSL.  All
** MemSL data structures are in the popular ADT format.
** 
** The C++ version has the following class interface:
** 
**           template <class USRDEF>
**           class WBPHeap
**             {
**               // private members omitted
**
**               protected:
**                 virtual int Lock(void *);
**                 virtual int Unlock(void *);
**                 virtual int RLock(void *);
**                 virtual int RUnlock(void *);
**                 virtual int Priority(void *, USRDEF, USRDEF);
**                 virtual int Compare(void *, USRDEF, USRDEF);
**                 virtual void Delete(void *, USRDEF);
**                 virtual void Execute(void *, USRDEF);
**
**               public:
**                 WBPHeap(unsigned int);
**                 ~WBPHeap();
**
**                 // by default the above virtual functions call
**                 // the user-defined call back functions defined
**                 // with the following functions.
**                 void LockF(void *, int (*)(void *));
**                 void UnlockF(void *, int (*)(void *));
**                 void RLockF(void *, int (*)(void *));
**                 void RUnlockF(void *, int (*)(void *));
**                 void PriorityF(void *, int (*)(void *, USRDEF, USRDEF));
**                 void CompareF(void *, int (*)(void *, USRDEF, USRDEF));
**                 void DeleteF(void *, void (*)(void *, USRDEF));
**                 void ExecuteF(void *, void (*)(void *, USRDEF));
**
**                 int Add(USRDEF);
**                 int operator+=(USRDEF item) { return(Add(item)); }
**                 int Delete(USRDEF, int delitem = 1);
**                 int operator-=(USRDEF item) { return(Delete(item)); }
**                 void Clear(int delitem = 1);
**                 USRDEF Search(USRDEF);
**                 USRDEF IsMember(USRDEF item) { return(Search(item)); }
**                 USRDEF Get();
**                 USRDEF operator--(int) { return(Get()); }
**                 USRDEF Peek();
**                 void Execute();
**                 long NumItems();
**                 int IsEmpty();
**             };
** 
** The above is a typical example of the C++ classes as defined in the
** MemSL, with all C++ abilities used to their fullest.
** 
** Now on to the macros below.  These macros provide all the functionality
** required with a priority heap.  To use the macros you will first need
** to set up a few things.  The following is what you will need:
** 
** an array - The array can be of any data type that allows assignment
**            in the following way:
** 
**                 array[index] = array[index+n]
** 
**            example:
**                struct address_rec
**                  {
**                    char name[30],
**                         address1[30],
**                         address2[30],
**                         address3[30],
**                         address4[30],
**                         phone1[20],
**                         phone2[20];
**                    int priority;
**                  } *addresses[100];
** 
** the size of the array - Value integer.  This will be used frequently and
**                         must not change.
** 
**                         example: int maxheap = 100;
** 
** the current heapsize - Value integer.  This value will change but should
**                        not be changed outside of the macros.
** 
**                        example: int heapsize = 0;
** 
** priority function - A function used to evaluate the priority of two
**                     items.  This function is similar to a comparison
**                     function and should return 1, 0, -1 depending on
**                     whether the priority of item1 is greater than item2,
**                     equal to item2 or less than item2.  By returning
**                     the values as stated, the priority heap will return
**                     the smallest value in the heap during the retrieval
**                     process (also called a minimum priority queue).
**                     By reversing the values returned by the priority
**                     function you can change the priority heap into a
**                     maximum priority queue.
** 
**                     example:
**
**                     int Priority(void *notused,
**                                  struct address_rec *address1,
**                                  struct address_rec *address2)
**                       {
**                         // For minimum priority
**                         return(address1->priority - address2->priority);
**
**                         // For maximum priority
**                         // return(address2->priority - address1->priority);
**                       }
** 
** priority parameter - This is a user-defined data item and is optional.
**                      Notice in the above example there is defined a void
**                      pointer named "notused".  This is the priority
**                      parameter and simply allows additional data to be
**                      passed to the function.
** 
** temporary item - The temporary item is used to temporarily store one
**                  item in the array while swapping items.  The temporary
**                  item should be the same type and size of one array item.
** 
**                  example: struct address_rec *tmpitem;
** 
** To add items use WBHEAP_ADD.  You may also want to check for available
** space with WBHEAP_OPENENTRIES as there is no other way of telling
** whether the item will be added to the heap.
** 
** To retrieve an item from the heap use WBHEAP_GETITEM.  The WBHEAP_GETITEM
** macro does not change the array, it simply returns the item at the given
** index.  With a minimum priority queue the item at array[0] is
** guaranteed to be the lowest in priority and the reverse is true with
** a maximum priority queue.  You may also want to check that there is
** a valid item at the given index with WBHEAP_VALIDITEM.
** 
** To remove an item from the heap use WBHEAP_DELETE.  WBHEAP_DELETE works
** the same as WBHEAP_GETITEM with the exception that the array is adjusted
** to remove the entry.
** 
** example:
**        struct address_rec *addresses[100];
**        int maxheap = 100;
**        int heapsize = 0;
** 
**        int AddAddressRec(char *name,
**                          char *address1,
**                          char *address2,
**                          char *address3,
**                          char *address4,
**                          char *phone1,
**                          char *phone2,
**                          int priority)
**          {
**            struct address_rec *address;
** 
**            // Allocate and assign the elements of the address record.
**
**            if (WBHEAP_OPENENTRIES(heapsize,maxheap))
**              if ((address = malloc(sizeof(struct address_rec)) != NULL)
**                {
**                  // Set the elements
**
**                  strcpy(address->name,name);
**                  strcpy(address->address1,address1);
**                  strcpy(address->address2,address2);
**                  strcpy(address->address3,address3);
**                  strcpy(address->address4,address4);
**                  strcpy(address->phone1,phone1);
**                  strcpy(address->phone2,phone2);
**                  address->priority = priority;
** 
**                  WBHEAP_ADD(addresses,heapsize,maxheap,
**                             Priority,NULL,address);
** 
**                  return(1);
**                }
**            return(0);
**          }
** 
**        struct address_rec *GetFirstAddress(void)
**          {
**            struct address_rec *ritem = NULL, *tmpitem;
** 
**            if (WBHEAP_VALIDITEM(heapsize,0))
**              {
**                ritem = WBHEAP_GETITEM(addresses,0);
**                WBHEAP_DELETE(addresses,heapsize,Priority,NULL,tmpitem,0);
**              }
** 
**            return(ritem);
**          }
** 
\*************************************************************************/


/*************************************************************************\
** WBHEAP_OPENENTRIES
**
** Syntax:
**        unsigned int WBHEAP_OPENENTRIES(unsigned int heapsize,
**                                        unsigned int maxheap);
**
** Arguments:
**            heapsize is the current number of items in the heap.
**
**            maxheap is the maximum number of items allowed in the heap.
**
** Description:
**              This macro is used to test whether there is room for
**              adding items.
**
** Return Value:
**               This macro returns maxheap - heapsize.
\*************************************************************************/

#define WBHEAP_OPENENTRIES(heapsize, maxheap) heapsize < maxheap ? maxheap - heapsize : 0


/*************************************************************************\
** WBHEAP_ADD
**
** Syntax:
**         void WBHEAP_ADD(any_assignable_type array,
**                         unsigned int heapsize,
**                         unsigned int maxheap,
**                         unsigned int (*prifunct)(void *,
**                                                  any_assignable_type item1,
**                                                  any_assignable_type item2),
**                         void *priparam,
**                         any_assignable_type item);
**
** Arguments:
**            array is the array that will be used to store the items.
**            For the most part any_assignable_type indicates that the
**            array will be an array of pointers to some type.
**
**            heapsize is the current size of the heap (number of items).
**
**            maxheap is the maximum number of items that can be stored
**            in the array.
**
**            prifunct is the priority function, which must accept the
**            arguments of its own.  The first argument is the next
**            argument described and is simply a means for passing
**            additional information to the function and is not used
**            by the heap macros.  The second and third arguments are
**            elements of the array and must be assignable.
**
**            priparam is a user-definable item that will be passed to
**            the priority function during each call.
**
**            item is a single item being added to the array and must
**            also be assignable.
**
** Description:
**              This macro adds items to a priority heap.  The heap
**              should be check to make sure there is room for the
**              new item (see WBHEAP_OPENENTRIES) as there is no
**              other way to know whether the item was added or not.
**
** Return Value:
**               No return value.
\*************************************************************************/

#define WBHEAP_ADD(heap, heapsize, maxheap, prifunct, priparam, item) \
   {                                                                  \
     unsigned int mi_index;                                           \
                                      \
     if (heapsize < maxheap)                                          \
       {                                                              \
     mi_index = ++heapsize;                                       \
     while (mi_index > 1 && (*prifunct)(priparam,item,heap[(mi_index/2)-1]) > 0) \
       {                                                          \
         heap[mi_index-1] = heap[(mi_index/2)-1];                 \
         mi_index /= 2;                                           \
       }                                                          \
     heap[mi_index-1] = item;                                     \
       }                                                              \
   }


/*************************************************************************\
** WBHEAP_DELETE
**
** Syntax:
**         void WBHEAP_DELETE(any_assignable_type array,
**                           unsigned int heapsize,
**                           unsigned int (*prifunct)(void *,
**                                              any_assignable_type item1,
**                                              any_assignable_type item2),
**                           void *priparam,
**                           any_assignable_type tmpitem,
**                           unsigned int in_dex);
**
** Arguments:
**            array is the array that will be used to store the items.
**            For the most part any_assignable_type indicates that the
**            array will be an array of pointers to some type.
**
**            heapsize is the current size of the heap (number of items).
**
**            prifunct is the priority function, which must accept the
**            arguments of its own.  The first argument is the next
**            argument described and is simply a means for passing
**            additional information to the function and is not used
**            by the heap macros.  The second and third arguments are
**            elements of the array and must be assignable.
**
**            priparam is a user-definable item that will be passed to
**            the priority function during each call.
**
**            tmpitem is any type that can be used to temporarily store
**            a value in the array.
**
**            in_dex is the array index being deleted.
**
** Description:
**              This macro is used to remove items from the priority
**              heap followed by an adjustment of the heap.
**
** Return Value:
**               No return value.
\*************************************************************************/

#define WBHEAP_DELETE(heap, heapsize, prifunct, priparam, tmpitem, in_dex) \
   {                                                                  \
     unsigned int mi_tmpindex = in_dex + 1, mi_left, mi_right, mi_largest; \
                                      \
     if (heapsize)                                                    \
       {                                                              \
     heap[in_dex] = heap[--heapsize];                             \
     do                                                           \
       {                                                          \
         mi_left = mi_tmpindex*2;                                 \
         mi_right = mi_tmpindex*2+1;                              \
         if (mi_right <= heapsize                                 \
           && (*prifunct)(priparam,heap[mi_right-1],heap[mi_tmpindex-1]) > 0) \
           mi_largest = mi_right;                                 \
         else                                                     \
           mi_largest = mi_tmpindex;                              \
         if (mi_left <= heapsize                                  \
           && (*prifunct)(priparam,heap[mi_left-1],heap[mi_largest-1]) > 0) \
           mi_largest = mi_left;                                  \
             if (mi_largest != mi_tmpindex)                           \
               {                                                      \
                 tmpitem = heap[mi_tmpindex-1];                       \
                 heap[mi_tmpindex-1] = heap[mi_largest-1];            \
                 heap[mi_largest-1] = tmpitem;                        \
                 mi_tmpindex = mi_largest;                            \
           }                                                      \
             else mi_tmpindex = 0;                                    \
           }                                                          \
         while (mi_tmpindex);                                         \
       }                                                              \
   }


/*************************************************************************\
** WBHEAP_VALIDITEM
**
** Syntax:
**         int WBHEAP_VALIDITEM(unsigned int heapsize,
**                              unsigned int index);
**
** Arguments:
**            heapsize is the current size of the heap (number of items).
**
**            index is the offset being checked.
**
** Description:
**              This macro checks if the index is a valid index with in
**              the heap.
**
** Return Value:
**               Returns 1 if the index is valid.  Otherwise returns 0.
\*************************************************************************/

#define WBHEAP_VALIDITEM(heapsize, index) index < heapsize ? 1 : 0


/*************************************************************************\
** WBHEAP_GETITEM
**
** Syntax:
**         any_assignable_type WBHEAP_GETITEM(any_assignable_type array,
**                                            unsigned int index);
**
** Arguments:
**            array is the array that will be used to store the items.
**            For the most part any_assignable_type indicates that the
**            array will be an array of pointers to some type.
**
**            index is the array index being retrieved.
**
** Description:
**              This macro simply returns the value from the array
**              at index.
**
** Return Value:
**               Returns the value from the array at index.
\*************************************************************************/

#define WBHEAP_GETITEM(heap, index) heap[index]



